import plateo.tools as tools
import pytest

def invert_sublists(l):
    return [sl[::-1] for sl in l]


@pytest.mark.parametrize("num_wells, expected", [(48, (6, 8)),
                                                 (96, (8, 12)),
                                                 (384, (16, 24)),
                                                 (1536, (32, 48))])
def test_compute_rows_columns(num_wells, expected):
    assert tools.compute_rows_columns(num_wells) == expected


rowname_data = [
    ("A", 1),
    ("E", 5),
    ("AA", 27),
    ("AE", 31)
]

@pytest.mark.parametrize("rowname, expected", rowname_data)
def test_rowname_to_number(rowname, expected):
    assert tools.rowname_to_number(rowname) == expected


@pytest.mark.parametrize("number, expected", invert_sublists(rowname_data))
def test_number_to_rowname(number, expected):
    assert tools.number_to_rowname(number) == expected


coordinates_data = [
    ("A1", (1, 1)),
    ("C2", (3, 2)),
    ("C04", (3, 4)),
    ("H11", (8, 11)),
    ("AA7", (27, 7)),
    ("AC07", (29, 7))
]

@pytest.mark.parametrize("wellname, expected", coordinates_data)
def test_wellname_to_coordinates(wellname, expected):
    assert tools.wellname_to_coordinates(wellname) == expected


coord_to_name_data = [
    ((1, 1), "A1"),
    ((3, 2), "C2"),
    ((3, 4), "C4"),
    ((8, 11), "H11"),
    ((27, 7), "AA7"),
    ((29, 7), "AC7")
]


@pytest.mark.parametrize("coords, expected", coord_to_name_data)
def test_coordinates_to_wellname(coords, expected):
    assert tools.coordinates_to_wellname(coords) == expected

wellname_data = [
    ("A5", 96, "row", 5),
    ("A5", 96, "column", 33),
    ("C6", 96, "row", 30),
    ("C6", 96, "column", 43),
    ("C6", 384, "row", 54),
    ("C6", 384, "column", 83)
]
inverted_wellname_data = [[s[-1], s[1], s[2], s[0]] for s in wellname_data]


@pytest.mark.parametrize("wellname, nwells, direction, expected",
                         wellname_data)
def test_wellname_to_index(wellname, nwells, direction, expected):
    assert tools.wellname_to_index(wellname, nwells, direction) == expected


@pytest.mark.parametrize("index, num_wells, direction, expected",
                         inverted_wellname_data)
def test_index_to_wellname(index, num_wells, direction, expected):
    assert tools.index_to_wellname(index, num_wells, direction) == expected

shift_data = [
    ("A1", 0, 0, "A1"),
    ("A1", 0, 3, "A4"),
    ("A1", 3, 0, "D1"),
    ("A1", 3, 3, "D4"),
    ("Z16", 3, 3, "AC19")
]


@pytest.mark.parametrize("wellname, row_shift, column_shift, expected",
                         shift_data)
def test_shift_wellname(wellname, row_shift, column_shift, expected):
    assert tools.shift_wellname(wellname, row_shift, column_shift) == expected


infer_size_data = [
    [("A1", "D5"), 96],
    [("C10", "E5", "H12"), 96],
    [("C10", "I5"), 384],
    [("C14", "A5"), 384],
    [("AA1", "D5", "H12"), 1536],
    [("A1", "D30"), 1536],

]


@pytest.mark.parametrize("wellnames, expected", infer_size_data)
def test_infer_plate_size_from_wellnames(wellnames, expected):
    """Return the first of 96, 384, or 1536, to contain all wellnames."""
    assert tools.infer_plate_size_from_wellnames(wellnames) == expected



from ..PickList import PickList
from ..Well import TransferError
from ..tools import round_at, did_you_mean


class AssemblyPicklistGenerator:

    def __init__(self, part_mol=None, part_l=None, part_g=None,
                 complement_to=None, buffer_volume=0, volume_rounding=None,
                 minimal_dispense_volume=0):

        self.part_mol = part_mol
        self.part_l = part_l
        self.part_g = part_g
        self.buffer_volume = buffer_volume
        self.complement_to = complement_to
        self.volume_rounding = volume_rounding
        self.minimal_dispense_volume = minimal_dispense_volume

    def make_picklist(self, assembly_plan, source_wells, destination_wells,
                      buffer_well=None, complement_well=None):
        source_wells = list(source_wells)
        destination_wells = list(destination_wells)
        destination_wells = destination_wells[:len(assembly_plan.assemblies)]

        part_wells = {}
        duplicates = {}
        for well in source_wells:
            well_part = self.get_part_from_well(well)
            if well_part in part_wells:
                if well_part not in duplicates:
                    duplicates[well_part] = []
                duplicates[well_part].append(well.name)
                previous_well = part_wells[well_part]
                previous_concentration = previous_well.content.concentration()
                if well.content.concentration() > previous_concentration:
                    part_wells[well_part] = well
            else:
                part_wells[well_part] = well

        required_parts = set(assembly_plan.all_parts_used())
        missing_parts = required_parts.difference(set(part_wells))
        if missing_parts:
            return None, {'missing_parts': {
                part: {
                    'featured_in': assembly_plan.assemblies_featuring(part),
                    'did_you_mean': [
                        (name, str(part_wells[name]))
                        for name in did_you_mean(
                            part, part_wells, min_score=50)
                    ]
                }
                for part in missing_parts
            }}

        picklist = PickList()
        iterator = zip(assembly_plan.assemblies.items(), destination_wells)
        for ((construct_name, parts), destination_well) in iterator:
            destination_well.data.construct = construct_name
            for part in parts:
                source_well = part_wells[part]
                volume = self.volume_from_well(source_well,
                                               assembly_plan.parts_data)
                volume = round_at(volume, self.volume_rounding)
                volume = max(volume, self.minimal_dispense_volume)
                picklist.add_transfer(source_well, destination_well,
                                      volume=volume)

        wells_over_desired_volume = []
        if self.complement_to is not None:
            if complement_well is None:
                for label in 'COMPLEMENT', 'WATER':
                    complement_well = part_wells.get(label, None)
                if complement_well is None:
                    raise ValueError("Could not identify the complement well ."
                                     " in the plate. One well should be marked"
                                     " as containing WATER or COMPLEMENT.")
            if complement_well.is_empty:
                raise TransferError("Well with COMPLEMENT/WATER is empty: %s."
                                    % (complement_well))
            for well in destination_wells:
                to_well = picklist.restricted_to(destination_well=well)
                total_transfer_volume = to_well.total_transfered_volume()
                complement_volume = (self.complement_to -
                                     total_transfer_volume -
                                     self.buffer_volume)
                if complement_volume < 0:
                    wells_over_desired_volume.append(well)
                else:
                    picklist.add_transfer(source_well=complement_well,
                                          destination_well=well,
                                          volume=complement_volume)
        if self.buffer_volume:
            if buffer_well is None:
                buffer_well = part_wells.get('BUFFER', None)
            if buffer_well.is_empty:
                raise TransferError("Empty buffer well: %s." % (buffer_well))
            buffer_volume = round_at(self.buffer_volume, self.volume_rounding)
            for well in destination_wells:
                picklist.add_transfer(buffer_well, well, volume=buffer_volume)

        return picklist, {
            'wells_over_desired_volume': wells_over_desired_volume,
            'duplicates': {
                part_name: {
                    'wells': wells,
                    'selected': part_wells[part_name]
                }
                for part_name, wells in duplicates.items()
            }
        }

    @staticmethod
    def get_part_from_well(well):
        if 'part' in well.data:
            return well.data['part']
        else:
            return well.content.components_as_string().strip(" ")

    def get_part_molar_weight(self, part_data):
        """Returns the molar weight of the sequence in g/m.
        http://cels.uri.edu/gsc/cndna.html
        """
        if 'size' in part_data:
            size = part_data['size']
        else:
            size = len(part_data['record'])
        return 650 * size

    def volume_from_well(self, well, parts_data):
        part_mol = self.part_mol
        part_g = self.part_g
        part_l = self.part_l
        if part_mol is not None:
            part = self.get_part_from_well(well)
            molar_weight = self.get_part_molar_weight(parts_data[part])
            part_g = molar_weight * part_mol
        if part_g is not None:
            part_l = part_g / well.content.concentration()
        return part_l



from copy import deepcopy

try:
    from bokeh.plotting import figure, ColumnDataSource
    from bokeh.models import (
        Range1d, TapTool, HoverTool, OpenURL
    )
    BOKEH_AVAILABLE = True
except ImportError:
    BOKEH_AVAILABLE = False

from ..tools import (
    dicts_to_columns,
    wellname_to_coordinates,
    number_to_rowname
)
import numpy as np


def plate_to_bokeh_plot(plate, hover_data=(), well_to_html=None,
                        well_color_function=None):
    """Return an interactive bokeh plot of the plate.

    Hovering the wells displays some data on the wells.

    Parameters
    ----------

    plate
      The plate to be converted

    hover_data
      list or tuple of all fields from the well's data that should be
      displayed when hovering a well

    well_to_html
      Html that sould be displayed when hovering a well (only works if
      hover_data is left empty).

    well_color_function
      A function well=> #a103ba associating a color to fill each well
    """

    if not BOKEH_AVAILABLE:
        raise ImportError(
            "Function plate_to_bokeh_plot requires Bokeh installed")
    wells = deepcopy(plate.wells)

    if well_color_function is None:
        def well_color_function(well):
            return "#fff" if well.content == {} else "#aaa"

    if hover_data != ():
        def well_to_html(well):
            return "\n".join(
                [well.name] + [
                    "%s: %s" % (field, data.get(field, ""))
                    for field in hover_data
                ]
            )
    elif well_to_html is None:
        well_to_html = lambda well: well.name

    for name, well in wells.items():
        data = {
            field: info
            for field, info in data.items()
            if not isinstance(info, dict)
        }

    n_rows, n_columns = plate.num_rows, plate.num_columns
    p = figure(plot_width=600, plot_height=400,
               tools="box_zoom,reset,tap,save",
               x_range=Range1d(0, n_columns + 2),
               y_range=Range1d(0, n_rows + 2),
               responsive=True)

    placeholder_wells = p.circle(
        x="x", y="y", radius=0.3, fill_color=None,
        line_width=1, line_color="gray", name="placeholder_well",
        source=ColumnDataSource(dicts_to_columns([
            {
                "x": x + 2,
                "y": y + 1
            }
            for y in range(n_rows)
            for x in range(n_columns)
        ]))
    )

    dicts = []
    for name, well in wells.items():
        row, column = wellname_to_coordinates(name)
        well_infos = {
            "display_color": well_color_function(well),
            "bokeh_x": column + 1,
            "bokeh_y": n_rows + 1 - row,
            "html_content": well_to_html(well)
        }
        #well_infos.update(data)
        dicts.append(well_infos)

    actual_wells = p.circle(
        x="bokeh_x", y="bokeh_y", radius=0.3, fill_color='display_color',
        line_width=1, line_color="black", name="well",
        source=ColumnDataSource(dicts_to_columns(dicts))
    )

    text = p.text(
        x="x", y="y", text="text", text_baseline="middle",
        text_align="center", text_font_size="%dpx" % (
          0.8 * 144 / int(np.round(np.sqrt(plate.num_wells / 6)))),
        source=ColumnDataSource(dicts_to_columns([
            {
                "text": number_to_rowname(i + 1),
                "x": 1,
                "y": n_rows - i
            }
            for i in range(n_rows)
        ] + [
            {
                "text": str(i + 1),
                "x": i + 2,
                "y": n_rows + 1
            }
            for i in range(n_columns)
        ]))
    )
    if well_to_html is not None:
        tooltips = "@html_content"
    else:
        tooltips = "<u><b>@well_name</b></u><br/>" + " ".join(
            ["@%s" % field for field in hover_data]
        )
    hover = HoverTool(names=["well"], tooltips=tooltips)
    p.add_tools(hover)

    if any(("url" in well) for well in wells):
        taptool = p.select(type=TapTool)
        taptool.callback = OpenURL(url="@url")

    p.toolbar.logo = None
    p.yaxis.visible = False
    p.xaxis.visible = False
    p.xgrid.grid_line_color = None
    p.ygrid.grid_line_color = None
    return p



"""Parsers for getting plate or picklist informations from data files."""

from .plate_from_aati_fragment_analyzer import (
    plate_from_aati_fragment_analyzer_peaktable,
    plate_from_aati_fragment_analyzer_zip
)

from .picklist_from_labcyte_echo_logfile import \
    picklist_from_labcyte_echo_logfile

from .picklist_from_tecan_evo_picklist_file import \
    picklist_from_tecan_evo_picklist_file

from .plate_from_nanodrop_xml_file import plate_from_nanodrop_xml_file

from .plate_volumes_from_labcyte_echo_files import (
    plate_volumes_from_labcyte_echo_logfile,
    plate_volumes_from_labcyte_echo_survey,
)

from .picklist_from_labcyte_echo_picklist_file import (
    picklist_from_labcyte_echo_picklist_file
)

from .plate_from_tables import (
    plate_from_platemap_spreadsheet,
    plate_from_list_spreadsheet,
    plate_from_dataframe,
    plate_from_content_spreadsheet
)

from .plate_from_geneart_shipment_layout_sheet import (
    plates_from_geneart_shipment_layout_sheet
)


"""Module to produce animations of the picklist (one image per transfer).


**Organization:**

.. mermaid::

    graph TD
    pt[plot_transfer]
    lsatp[list_source_and_target_plates]
    mtf[make_transfer_figure]
    a[animate]
    wpdf[write_pdf]
    mf[message_function]
    mplftnpim[mplfig_to_npimage]
    pp[plot_plates]

    lsatp --> pp
    pp --> pt
    pt -->mtf
    mf --> mtf
    mtf --> wpdf
    mplftnpim --> a
    mtf --> a
"""


from matplotlib.patches import ConnectionPatch
from proglog import ProgressBarLogger, TqdmProgressBarLogger
from matplotlib.backends.backend_agg import FigureCanvasAgg
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
import imageio


class PicklistAnimator:
    """A picklist animator !

    Parameters
    ----------
    plate_plotters
      A dictionnary ``{plate_name: plotter_to_use_for_that_plate}``
      Where the plotter is a Plateo PlatePlotter.

    message_function
      A function ``(picklist, transfer) => string`` taking as input a Plateo
      Picklist and Transfer objects.

    logger
      Either None for no logger, 'bar' for a progress bar logger, or any custom
      Proglog progress bar logger.


    Examples
    --------

    >>> animator = PicklistAnimator(
    >>>     plate_plotters={
    >>>         'Source_1': success_plotter,
    >>>         'Source_2': success_plotter,
    >>>         'Target': volume_plotter
    >>>     },
    >>>     logger=TqdmProgressBarLogger(bars=['transfers'])
    >>> )
    >>> animator.animate(picklist, "test.mp4", fps=3)


    """

    def __init__(self, plate_plotters, message_function=None,
                 plate_figure_size=(8, 6), logger='bar'):
        """Initialize."""
        self.plate_plotters = plate_plotters
        if logger is None:
            logger = ProgressBarLogger()
        elif logger == 'bar':
            logger = TqdmProgressBarLogger(bars=['transfers'])
        self.logger = logger
        self.message_function = message_function
        self.plate_figure_size = plate_figure_size

    @staticmethod
    def mplfig_to_npimage(fig):
        """ Converts a matplotlib figure to a RGB array"""
        #  only the Agg backend now supports the tostring_rgb function
        canvas = FigureCanvasAgg(fig)
        canvas.draw()
        l, b, w, h = canvas.figure.bbox.bounds
        w, h = int(w), int(h)
        buf = canvas.tostring_rgb()
        image = np.fromstring(buf, dtype=np.uint8)
        return image.reshape(h, w, 3)

    @staticmethod
    def list_source_and_target_plates(picklist):
        """Return two lists: plates with resp. source wells and target wells.

        In the final animation, source plates will be on the left and target
        plates on the right.
        """
        source_plates = []
        target_plates = []
        seen_plates = set()
        for _transfer in picklist.transfers_list:
            source_plate = _transfer.source_well.plate
            target_plate = _transfer.destination_well.plate
            if source_plate not in seen_plates:
                source_plates.append(source_plate)
            if target_plate not in seen_plates:
                target_plates.append(target_plate)
            seen_plates.add(source_plate)
            seen_plates.add(target_plate)
        return source_plates, target_plates

    def plot_plates(self, source_plates, target_plates, axes=None):
        """Plot all plates, sources on the left, targets on the right.

        Parameters
        ----------

        source_plates
          List of plates with (at least mostly) source wells.

        target_plates
          List of plates with (at least mostly) destination wells.

        axes
          (left, right): Two lists of matplotlib axes, one for the source
          plates on the left, one for the destination plates on the right.

        """
        if axes is None:
            max_rows = max(len(source_plates), len(target_plates))
            w, h = self.plate_figure_size
            fig, axes = plt.subplots(max_rows, 2,
                                     figsize=(2 * w, h * max_rows))
            if max_rows == 1:
                axes = np.array([axes])
            for ax in axes.flatten():
                ax.axis('off')
                ax.set_aspect('equal')
        else:
            fig = axes[0][0].figure
            for ax in axes.flatten():
                ax.clear()
                ax.axis('off')
                ax.set_aspect('equal')

        axes_dict = {}
        for c, plates in enumerate([target_plates, source_plates]):
            for i, plate in enumerate(plates):
                column = 1 - c
                ax = axes[i, column]
                ax.set_title(plate.name, fontsize=20)
                self.plate_plotters[plate.name].plot_plate(plate, ax=ax)
                axes_dict[plate.name] = ax
        for i in range(5):
            fig.tight_layout()
        return fig, axes, axes_dict

    def plot_transfer(self, transfer, source_plates, target_plates, axes=None):
        """Plot the plates and add and arrow for the transfer"""
        fig, axes, axes_dict = self.plot_plates(source_plates, target_plates,
                                                axes=axes)

        source_well = transfer.source_well
        source_plate = source_well.plate
        source_ax = axes_dict[source_plate.name]
        x, y = source_well.coordinates[::-1]
        source_coord = x, source_plate.num_rows - y + 1

        target_well = transfer.destination_well
        target_plate = target_well.plate
        target_ax = axes_dict[target_plate.name]
        x, y = target_well.coordinates[::-1]
        target_coord = x, target_plate.num_rows - y + 1

        target_ax.set_zorder(source_ax.get_zorder() - 1)
        arrow = ConnectionPatch(
            xyA=source_coord, xyB=target_coord,
            coordsA="data", coordsB="data",
            axesA=source_ax, axesB=target_ax,
            shrinkB=5.0, lw=2,
            facecolor='black',
            arrowstyle="wedge", zorder=1000)
        source_ax.add_artist(arrow)
        return fig, axes

    def _make_transfer_figure(self, picklist, transfer=None, axes=None):
        sources, targets = self.list_source_and_target_plates(picklist)
        if transfer is not None:
            index = self.logger.bars['transfers']['index']
            self.logger(transfers__index=index + 1)
            fig, axes = self.plot_transfer(transfer, sources, targets,
                                           axes=axes)
        else:
            fig, axes, _ = self.plot_plates(sources, targets)

        if self.message_function is not None:
            message = self.message_function(picklist, transfer)
            if message is not None:
                fig.subplots_adjust(top=0.8)
                fig.suptitle(message, fontsize=12)
        return fig, axes

    def animate(self, picklist, target, dpi=160, fps=5):
        """Create a movie of the picklist.

        Parameters
        ----------
        picklist
          Picklist to be animated

        target
          File path or file-like object where to write the movie.

        dpi
          Resolution of the image. The larger the dpi, the larger the image.

        fps
          Frames per second
        """
        self.logger(transfers__total=len(picklist.transfers_list))
        writer = imageio.get_writer(target, fps=fps)
        fig, axes = self._make_transfer_figure(picklist)

        def make_frame(picklist, transfer):
            fig, _ = self._make_transfer_figure(picklist, transfer, axes=axes)
            fig.dpi = dpi
            writer.append_data(self.mplfig_to_npimage(fig))
        picklist.execute(inplace=False, callback_function=make_frame)
        writer.close()
        plt.close(fig)

    def write_pdf(self, picklist, target):
        """Create a multi-page PDF animation of the picklist.

        Parameters
        ----------


        """
        self.logger(transfers__total=len(picklist.transfers_list))

        with PdfPages(target) as pdf:
            fig, axes = self._make_transfer_figure(picklist)

            def make_frame(picklist, transfer):
                fig, _ = self._make_transfer_figure(picklist, transfer, axes=axes)
                pdf.savefig(fig, bbox_inches="tight")
                plt.close(fig)
            picklist.execute(inplace=False, callback_function=make_frame)



import pandas

from plateo.tools import index_to_wellname
from plateo.parsers.file_parsers import parse_excel_xml
from plateo.parsers.plate_from_tables import plate_from_dataframe
import numpy as np


def parse_numeric(numeric_wannabe):
    try:
        return float(numeric_wannabe)
    except ValueError:
        return np.nan

def plate_from_nanodrop_xml_file(xml_file=None, xml_string=None, num_wells=96,
                                 direction="row"):
    """Return a plate with the DNA concentrations measured by the Nanodrop.

    Parameters
    ----------

    xml_file
      The xml file exported from the Nanodrop software. It should contain one
      measurement per well. An xml_string can be provided instead of this
      argument.

    xml_string
       The content of an xml file exported from the Nanodrop software. Can be
       provided instead of xml_file. Should contain one measurement per well.

    num_wells
      Number of wells in the Plate

    direction
      Either ``row`` or `column``. Order of the measurements on the plate.


    Returns
    -------

    A Plate where the wells have the following data fields:
    TODO: complete

    """

    table = parse_excel_xml(xml_file=xml_file, xml_string=xml_string)[0]
    table = [
        line for line in table
        if len(line) == len(table[0])
    ]
    dataframe = pandas.DataFrame(table[1:], columns=table[0])
    for column in dataframe.columns:
        dataframe[column] = pandas.to_numeric(dataframe[column],
                                              errors='ignore')

    def wellname(i):
        return index_to_wellname(int(i), num_wells, direction=direction)
    dataframe["wellname"] = dataframe["#"].apply(wellname)
    conc_label = [
        label
        for label in ["Nucleic Acid", "Nucleic Acid Conc."]
        if label in dataframe
    ][0]
    dataframe['concentration'] = [parse_numeric(e)
                                  for e in dataframe[conc_label]]
    return plate_from_dataframe(dataframe, num_wells=num_wells)



"""Classes to represent picklists and liquid transfers in general"""
from collections import OrderedDict
from copy import deepcopy
import json

import pandas

#import parsers
#import exporters
from .tools import compute_rows_columns, wellname_to_index, index_to_wellname



class Transfer:
    """A tranfer from a source to a destination

    Parameters
    ----------

    source_well
      A Well object representing the plate well from which to transfer

    destination_well
      A Well object representing the plate well to which to transfer.

    volume
      Volume to be transfered, expressed in liters.

    data
      A dict containing any useful information on the transfer, this
      information can be used later e.g. as parameters for the transfer
      when exporting a picklist.
    """
    def __init__(self, source_well, destination_well, volume, data=None):

        self.volume = volume
        self.source_well = source_well
        self.destination_well = destination_well
        self.data = data

    def to_plain_string(self):
        """Return "xx L from {source_well} into {dest_well}"."""
        return ("{self.volume:.02E}L from {self.source_well.plate.name} "
                "{self.source_well.name} into "
                "{self.destination_well.plate.name} "
                "{self.destination_well.name}").format(
                    self=self
        )

    def change_volume(self, new_volume):
        """Return a version of the transfer with a new volume."""
        return Transfer(source_well=self.source_well,
                        destination_well=self.destination_well,
                        volume=new_volume,
                        data=self.data)

    def __repr__(self):
        """Return "xx L from {source_well} into {dest_well}"."""
        return self.to_plain_string()


class PickList:
    """Representation of a list of well-to-well transfers.

    Parameters
    -----------

    transfers_list
      A list of Transfer objects that will be part of a same dispensing
      operation, in the order in which they are meant to be executed.

    data
      A dict with some infos on the picklist.

    """

    def __init__(self, transfers_list=(), data=None):

        self.transfers_list = list(transfers_list)
        self.data = {} if data is None else data

    def add_transfer(self, source_well=None, destination_well=None,
                     volume=None,  data=None, transfer=None):
        """Add a transfer to the picklist's tranfers list.

        You can either provide a ``Transfer`` object with the ``transfer``
        parameter, or the parameters


        """
        if transfer is None:
            transfer = Transfer(source_well=source_well,
                                destination_well=destination_well,
                                volume=volume,
                                data=data)
        self.transfers_list.append(transfer)

    def to_plain_string(self):
        """Return the list of transfers in human readable format"""
        return "\n".join(
            transfer.to_plain_string()
            for transfer in self.transfers_list
        )

    def to_plain_textfile(self, filename):
        """Write the picklist in a file in a human reable format."""
        with open(filename, "w+") as f:
            f.write(self.to_plain_string())

    def execute(self, content_field="content", inplace=True,
                callback_function=None):
        """Simulate the execution of the picklist"""

        if not inplace:
            all_plates = set(
                plate
                for transfer in self.transfers_list
                for plate in [transfer.source_well.plate,
                              transfer.destination_well.plate]
            )
            new_plates = {
                plate: deepcopy(plate)
                for plate in all_plates
            }

            new_transfer_list = []
            for transfer in self.transfers_list:
                new_source_plate = new_plates[transfer.source_well.plate]
                new_dest_plate = new_plates[transfer.destination_well.plate]
                new_source_well = new_source_plate.wells[
                    transfer.source_well.name]
                new_dest_well = new_dest_plate.wells[
                    transfer.destination_well.name]
                new_transfer_list.append(Transfer(
                    volume=transfer.volume,
                    source_well=new_source_well,
                    destination_well=new_dest_well
                ))

            new_picklist = PickList(transfers_list=new_transfer_list)
            new_picklist.execute(content_field=content_field, inplace=True,
                                 callback_function=callback_function)
            return new_plates

        else:
            for transfer in self.transfers_list:
                transfer.source_well.transfer_to_other_well(
                    destination_well=transfer.destination_well,
                    transfer_volume=transfer.volume)
                if callback_function is not None:
                    callback_function(self, transfer)

    def restricted_to(self, transfer_filter=None, source_well=None,
                      destination_well=None):
        """Return a version of the picklist restricted to transfers with the
        right source/destination well.

        You can provide ``source_well`` and ``destination_well`` or
        alternatively just a function ``transfer_filter`` with signature
        (transfer)=>True/False that will be used to filter out transfers
        (for which it returns false).

        """
        if transfer_filter is None:
            def transfer_filter(tr):
                source_well_is_ok = ((source_well is None) or
                                     (source_well == tr.source_well))
                dest_well_is_ok = ((destination_well is None) or
                                   (destination_well == tr.destination_well))
                return (source_well_is_ok and dest_well_is_ok)

        transfers = [tr for tr in self.transfers_list if transfer_filter(tr)]
        return PickList(transfers, data={"parent": self})

    def sorted_by(self, sorting_method="source_well"):
        """Return a new version of the picklist sorted by some parameter.

        The ``sorting_method`` is either the name of an attribute of the
        transfers, such as "source_well", or a function f(transfer) -> value.
        """
        if not hasattr(sorting_method, "__call__"):
            def sorting_method(transfer):
                return transfer.__dict__[sorting_method]
        return PickList(sorted(self.transfers_list, key=sorting_method),
                        data={"parent": self})

    def split_by(self, category, sort_key):
        """Split the picklist into a list of picklists, per category.

        The returned list if of the form [(cat, subpicklist)] where
        ``cat`` is the value of the category for all transfers in
        ``subpicklist``.

        The parameter ``category`` is either the name of a transfer attribute
        or a function f(transfer)=> value which is used to
        """
        if isinstance(category, str):
            str_category = category
            def category(t):
                return t.__dict__[str_category]
        categories = set([category(tr) for tr in self.transfers_list])
        return [
            (cat, self.restricted_to(lambda tr: category(tr) == cat))
            for cat in sorted(categories, key=sort_key)
        ]

    def total_transfered_volume(self):
        """Return the sum of all volumes from all transfers."""
        return sum([transfer.volume for transfer in self.transfers_list])

    @staticmethod
    def from_plates(source_plate, destination_plate, volume,
                    source_criterion=None,
                    destination_criterion=None, source_direction="row",
                    destination_direction="row"):
        """Create a PickList object based on plates and conditions.

        BROKEN due to changes in picklists. TODO: Fix.
        """

        if not hasattr(volume, "__call__"):
            constant_volume = volume
            volume = lambda source_well: constant_volume

        if source_criterion is None:
            source_criterion = lambda well: True
        if destination_criterion is None:
            destination_criterion = lambda well: True

        destination_wells = (
            well for well in destination_plate.iter_wells(
                direction=destination_direction)
            if destination_criterion(well)
        )
        transfers_list = []
        if isinstance(source_plate, (list, tuple)):
            source_wells = (
                p.iter_wells(direction=source_direction)
                for p in source_plate
            )
        else:
            source_wells = source_plate.iter_wells(direction=source_direction)
        for source_well in source_wells:
            if source_criterion(source_well):
                destination_well = next(destination_wells)
                transfers_list.append(
                    Transfer(
                        source_well=source_well,
                        destination_well=destination_well,
                        volume=volume(source_well)
                    )
                )

        return PickList(transfers_list)

    def enforce_maximum_dispense_volume(self, max_dispense_volume):
        """Return a new picklist were every too-large dispense is broken down
        into smaller dispenses."""
        transfers = []
        for trf in self.transfers_list:
            n_additional_dispense = int(trf.volume / max_dispense_volume)
            rest = trf.volume - n_additional_dispense * max_dispense_volume
            for i in range(n_additional_dispense):
                transfers.append(trf.change_volume(max_dispense_volume))
            if rest > 0:
                transfers.append(trf.change_volume(rest))
        return PickList(transfers_list=transfers)

    def __add__(self, other):
        return PickList(self.transfers_list + other.transfers_list)

    @staticmethod
    def merge_picklists(picklists_list):
        """Merge the list of picklists into a single picklist.

        The transfers in the final picklist are the concatenation of the
        tranfers in the different picklists, in the order in which they appear
        in the list.
        """
        return sum(picklists_list, PickList([]))



from ...version import __version__
from pdf_reports import ReportWriter
from sequenticon import sequenticon
import os

THIS_PATH = os.path.dirname(os.path.realpath(__file__))
TEMPLATES_PATH = os.path.join(THIS_PATH, "templates")

def template_path(name):
    return os.path.join(TEMPLATES_PATH, name)

report_writer = ReportWriter(
    plateo_logo_url=os.path.join(TEMPLATES_PATH, 'imgs', 'logo.png'),
    version=__version__,
    default_stylesheets=(template_path("default_style.css"),)
)



import pandas
import flametree
from ..containers import Plate96

def parts_ids_from_geneart_records_dir(folder):
    records_root = flametree.file_tree(folder)
    return {
        dir_._name.split("_")[0]: "_".join(dir_._name.split("_")[1:])
        for dir_ in records_root._dirs
    }

def plates_from_geneart_shipment_layout_sheet(filepath, parts_ids_dict=None,
                                              geneart_parts_dir=None):
    """Return a list of all plates (Plate96) in the shipment layout sheet.
    
    Example
    -------

    >>> plates = plates_from_geneart_shipment_layout_spreadsheet(
    >>>     filepath="2018AAMOBC_layout_shipment.xlsx",
    >>>     geneart_parts_dir="./tfcdownload(4).zip")
    >>> # Write the plates in reformatted format:
    >>> for plate in plates:
    >>>     plate_to_content_spreadsheet(plate, "./%s.xlsx" % plate.name)
    Parameters
    ----------
    
    filepath
      Path to the Geneart excel spreadsheet or a filelike.
    
    parts_ids_dict
      Optional. dictionnary {geneart_id: part_name}. If provided, the content
      of the wells will be indicated using your custom part_name, if not
      provided, the genart_id will be used.
    
    geneart_parts_dir
      Optional path to a folder, zip file, or Flametree-compatible dir, from
      which to read filenames like "18AFY2AC_p9_EGFP" from which the
      association between geneart IDs(here, 18AFY2AC) and your part names
      (p9_EGFP) will be read.

    
    """
    if geneart_parts_dir is not None:
        parts_ids_dict = parts_ids_from_geneart_records_dir(geneart_parts_dir)
    plates_data = pandas.read_excel(filepath, skiprows=2)
    plates_indices = sorted(set(plates_data.Plate))
    plates = {index: Plate96(name="Plate %d" % index)
              for index in plates_indices}
    for index, plate in plates.items():
        subdata = plates_data[plates_data.Plate == index]
        for i, row in subdata.iterrows():
            well = plate.wells[row.Pos]
            well.data = {
                k: row[k]
                for k in ('OC_Number', 'IDAuftrag', 'IDConstruct')
            }
            volume_microl = row['Volume shiped [µl]']
            volume_in_l = volume_microl * 1e-6
            quantity = volume_microl * row['Conzentration [µg/µl]'] * 1e-6
            part_label = row.IDAuftrag
            if parts_ids_dict is not None:
                part_label = parts_ids_dict[row.IDConstruct]
            else:
                part_label = row.IDConstruct
            well.add_content({part_label: quantity}, volume=volume_in_l)
    return [plate for i, plate in sorted(plates.items())]


from plateo.parsers import (
    plate_from_platemap_spreadsheet,
    plate_from_list_spreadsheet,
    plate_from_nanodrop_xml_file,
    plate_from_aati_fragment_analyzer_peaktable,
    plate_from_aati_fragment_analyzer_zip,
    plate_from_dataframe)


def test_plate_from_platemap_spreadsheet():
    plate_from_platemap_spreadsheet
    pass

def test_plate_from_list_spreadsheet():
    plate_from_list_spreadsheet
    pass

def test_plate_from_nanodrop_xml_file():
    plate_from_nanodrop_xml_file
    pass

def test_plate_from_dataframe():
    plate_from_dataframe
    pass


def test_plate_from_aati_fragment_analyzer_peaktable():
    plate_from_aati_fragment_analyzer_peaktable
    pass


def test_plate_from_aati_fragment_analyzer_zip():
    plate_from_aati_fragment_analyzer_zip
    pass



import textwrap

import numpy as np

try:
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    from mpl_toolkits.axes_grid.inset_locator import inset_axes
    MATPLOTLIB_AVAILABLE = True
except ImportError:
    MATPLOTLIB_AVAILABLE = False
from tqdm import tqdm

from ..tools import (
    compute_rows_columns,
    number_to_rowname
)

letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'


def draw_plate_layout(num_wells, ax):
    """Draw the plate's border, row letters, column numbers."""
    n_rows, n_columns = compute_rows_columns(num_wells)
    ax.axis("off")
    ax.add_patch(patches.Rectangle((0.5, 0.5), n_columns, n_rows, fill=False))
    for i in range(1, n_columns + 1):
        ax.text(i, 1.075 * n_rows, str(i), horizontalalignment="center")
    ax.set_ylim(0, n_rows + 2)
    for i in range(n_rows):
        ax.text(0.3, n_rows - i, number_to_rowname(i + 1),
                verticalalignment="center",
                horizontalalignment="right")
    ax.plot([n_rows], [n_columns])
    ax.set_xlim(0, n_columns + 0.6)


def place_inset_ax_in_data_coordinates(ax, bbox):
    """Return an ax inset in the given ax at the given bbox in
    data coordinates (bottom, left, width, height)"""

    bottom, left, width, height = bbox
    pixels_data_00 = ax.transData.transform([0, 0])
    pixels_data_wh = ax.transData.transform([width, height])
    iwidth, iheight = (pixels_data_wh - pixels_data_00) / ax.figure.dpi
    return inset_axes(
        ax, iwidth, iheight,
        loc=10,  # means "center"
        bbox_to_anchor=[bottom, left, width, height],
        bbox_transform=ax.transData
    )


class PlatePlotter:
    """Base class for all matplotlib-based plate plotters
    """

    def plot_plate(self, plate, ax=None, well_filter=None,
                   progress_bar=False, figsize=(10, 5)):
        """Plot the plate using Matplotlib.

        Parameters
        ----------

        plate
          The plate object to be plotted.

        ax
          The Matplotlib ax on which to plot. If none is provided, one will
          be created (and returned at the end)

        well_filter
          Function f(well)=>true/false. Wells returning false will not be
          considered.

        progress_bar
          If true, display a progress bar while plotting.

        figsize
          Size of the figure, width/height in inches.


        Returns
        -------

        ax, stats
          The matplotlib ax on which the figure was plotted, and the computed
          stats in format {wellname: well_stat}, (if relevant)
        """

        if not MATPLOTLIB_AVAILABLE:
            raise IOError("Install Matplotlib to be able to plot")
        if ax is None:
            fig, ax = plt.subplots(1, facecolor="white", figsize=figsize)
        if well_filter is None:
            well_filter = lambda well:  True

        draw_plate_layout(plate.num_wells, ax)

        stats = {}

        def progress(plate):
            if progress_bar:
                return tqdm(plate, total=plate.num_wells)
            else:
                return plate
        for well in progress(plate):
            if not well_filter(well):
                continue

            x, y = well.column, well.row
            y = plate.num_rows - y + 1

            stats[well.name] = self.plot_well(ax, x, y, well)

        self.post_process(ax, stats)
        return ax, stats

    def post_process(self, ax, stats):
        pass


class PlateColorsPlotter(PlatePlotter):
    """Plot a plate's well statistic with wells colored differently.

    Parameters
    ----------

    stat_function
      The function to be plotted, with signature (well) => value

    with_colorbar
      It true a colorbar giving a scale for color values will be plotted
      alongside the map

    """

    def __init__(self, stat_function, colormap=None, plot_colorbar=False,
                 well_radius='full', vmin=None, vmax=None, alpha=1.0,
                 edge_width=1):
        self.stat_function = stat_function
        self.colormap = colormap
        self.plot_colorbar = plot_colorbar
        self.well_radius = well_radius
        self.alpha = alpha
        self.vmin = vmin
        self.vmax = vmax
        self.edge_width = edge_width

    def plot_well(self, ax, x, y, well):
        return ((x, y), self.stat_function(well))

    def post_process(self, ax, stats):
        xy, stats_values = zip(*stats.values())
        xx, yy = zip(*xy)
        if self.well_radius == 'full':
            values = list(stats.values())
            colors = [
                v[1] for v in values if (v[1] is not None)
            ]
            depth = len(colors[0]) if hasattr(colors[0], '__iter__') else 1
            my, mx = max(yy), max(xx)
            grid = np.zeros((my, mx, depth) if depth > 1 else (my, mx))
            for (x, y), stat in stats.values():
                grid[y-1, x-1] = stat
            plot = ax.imshow(grid[::-1, :], alpha=self.alpha,
                             vmin=self.vmin, vmax=self.vmax,
                             cmap=self.colormap,
                             extent=[0.5, max(xx)+0.5, 0.5, max(yy)+0.5])


        else:
            plot = ax.scatter(xx, yy, s=self.well_radius, c=stats_values,
                              vmin=self.vmin, vmax=self.vmax,
                              linewidths=self.edge_width,
                              edgecolors='k',
                              alpha=self.alpha, cmap=self.colormap)
        if self.plot_colorbar:
            ax.figure.colorbar(plot)


class PlateTextPlotter(PlatePlotter):
    """Plot a plate's well statistic as text at the well's positions.

    Parameters
    ----------

    text_function
      A function (well) => text

    with_colorbar
      It true a colorbar giving a scale for color values will be plotted
      alongside the map
    """

    def __init__(self, text_function, line_length=None, fontdict=None):

        self.fontdict = {} if fontdict is None else fontdict
        self.text_function = text_function
        self.line_length = line_length

    def plot_well(self, ax, x, y, well):
        text = str(self.text_function(well))
        if self.line_length is not None:
            text = '\n'.join(textwrap.wrap(text, self.line_length))
        fontdict = self.fontdict
        if not isinstance(fontdict, dict):
            fontdict = fontdict(well)

        ax.text(x, y, str(text),
                fontdict=fontdict,
                horizontalalignment="center",
                verticalalignment="center")
        return ((x, y), text)


class PlateGraphsPlotter(PlatePlotter):
    """Plot a graph (for instance time series) for each well of the plate

    The result has the shape of a plate, with each well replaced by a graph.

    Parameters
    ----------

     plot_function
       A function f(well, well_ax) which plots something on the provided
       ``well_ax`` depending on the ``well``.

     subplot_size
       (width, height) of subplots, between 0 and 1 (1 meaning the graphs of
       the different wells will have virtually no margin between them)

    """

    def __init__(self, plot_function, subplot_size=(0.7, 0.7)):

        self.plot_function = plot_function
        self.subplot_size = subplot_size
        self.subplot_width = float(subplot_size[0])
        self.subplot_height = float(subplot_size[1])

    def plot_well(self, ax, x, y, well):
        bbox = (x - self.subplot_width / 2.0,
                y - self.subplot_height / 2.0,
                self.subplot_width,
                self.subplot_height)
        well_ax = place_inset_ax_in_data_coordinates(ax, bbox)
        self.plot_function(well, well_ax)
        return well_ax



"""This module implements the Base class for all plates.

See plateo.container for more specific plate subclasses, with
set number of wells, well format, etc.
"""
from collections import OrderedDict
import json
from .Well import Well
from .tools import (index_to_wellname, wellname_to_index,
                    coordinates_to_wellname, rowname_to_number,
                    replace_nans_in_dict)
from box import Box

class Plate:
    """Base class for all wells."""

    PlateWell = Well

    def __init__(self, name=None, wells_data=None,
                 data=None):

        self.name = name
        self.data = Box({} if data is None else data)
        self.wells_data = {} if wells_data is None else wells_data
        self.num_wells = self.num_rows * self.num_columns
        self.wells = Box()

        for row in range(1, self.num_rows + 1):
            for column in range(1, self.num_columns + 1):
                wellname = coordinates_to_wellname((row, column))
                data = self.wells_data.get(wellname, {})
                well = self.PlateWell(plate=self, row=row, column=column,
                                      name=wellname, data=data)
                self.wells[wellname] = well

    def __getitem__(self, k):
        """Return e.g. well A1's dict when calling `myplate['A1']`."""
        return self.wells[k]

    def merge_data_from(self, other_plate, overwrite=True):
        """Adds a new field `field_name` to the

        Note that `fun` can also return nothing and simply transform the wells.
        """
        for well in self:
            if well.name in other_plate.wells.keys():
                other_well = other_plate[well.name]
                other_data = other_well.data
                if not overwrite:
                    other_data = {k: v for (k, v) in other_data.items()
                                      if k not in well.data}
                well.data.update(other_data)

    def apply_to_wells(self, fun):
        """Run fun(well) for every `name:well` in `self.wells_dict`"""
        for well in self:
            fun(well)

    def compute_data_field(self, field_name, fun, ignore_none=False):
        for well in self:
            data = fun(well)
            if (data is not None) or (not ignore_none):
                well.data[field_name] = data

    def find_unique_well(self, content_includes=None, condition=None):
        if content_includes is not None:
            def condition(well):
                return (content_includes in well.content.quantities.keys())
        wells = [
            well
            for name, well in self.wells.items()
            if condition(well)
        ]
        if len(wells) > 1:
            raise ValueError("Query returned several wells: %s" % wells)
        elif len(wells) == 0:
            raise ValueError("No wells found matching the condition")
        return wells[0]

    def to_pretty_string(self, well_name, indent=2):
        return json.dumps(self[well_name], indent=indent)

    def list_well_data_fields(self):
        return sorted(list(set(
            field
            for well in self
            for field in well.data.keys()
        )))

    def wells_in_column(self, column_number):
        """Return the list of all wells of the plate in the given column."""
        # TODO: at some point, avoid iterating over all wells, make it smarter
        return [
            well for well in self
            if well.column == column_number
        ]

    def wells_in_row(self, row):
        """Return the list of all wells of the plate in the given row.

        The `row` can be either a row number (1,2,3) or row letter(s) (A,B,C).
        """
        if isinstance(row, str):
            row = rowname_to_number(row)
        return [
            well for well in self
            if well.row == row
        ]

    def wells_satisfying(self, condition):
        """

        Examples
        ---------
        >>> def condition(well):
        >>>     return well.volume > 50
        >>> for well in myplate.wells_satifying(condition):
        >>>     print( well.name )
        """
        return filter(condition, self.wells.values())

    def wells_grouped_by(self, data_field=None, key=None, sort_keys=False,
                         ignore_none=False, direction_of_occurence="row"):
        if key is None:
            def key(well):
                return well.data.get(data_field, None)
        dct = OrderedDict()
        for well in self.iter_wells(direction=direction_of_occurence):
            well_key = key(well)
            if well_key not in dct:
                dct[well_key] = [well]
            else:
                dct[well_key].append(well)
        if ignore_none:
            dct.pop(None, None)
        keys = dct.keys()
        if sort_keys:
            keys = sorted(keys)
        return [(k, dct[k]) for k in keys]

    def get_well_from_index(self, index, direction="row"):
        return self[self.index_to_wellname(index, direction=direction)]

    def well_at_index(self, index, direction="row"):
        return self[self.index_to_wellname(index, direction=direction)]

    def index_to_wellname(self, index, direction="row"):
        return index_to_wellname(index, self.num_wells, direction=direction)

    def wellname_to_index(self, wellname, direction="row"):
        return wellname_to_index(wellname, self.num_wells, direction=direction)

    def iter_wells(self, direction="row"):
        """Iter through the wells either by row or by column"""
        if direction == "row":
            return self.wells_sorted_by(lambda w: (w.row, w.column))
        else:
            return self.wells_sorted_by(lambda w: (w.column, w.row))

    def wells_sorted_by(self, sortkey):
        return (e for e in sorted(self.wells.values(), key=sortkey))
    
    def list_data_field_values(self, data_field, include_none=False):
        return list(set([
            w.data[data_field]
            for w in self.iter_wells()
            if data_field in w.data
            and (include_none or (w.data[data_field] is not None))
        ]))

    def last_nonempty_well(self, direction='row'):
        """Return the last non-empty well found when traversing the plate."""
        selected_well = None
        for well in self.iter_wells(direction=direction):
            if not well.is_empty:
                selected_well = well
        return selected_well

    def __iter__(self):
        """Allow to iter through the well dicts using `for well in myplate`"""
        return self.iter_wells()

    def to_dict(self, replace_nans_by='null'):
        dct = {
            "data": self.data,
            "wells": {
                well.name: well.to_dict()
                for well in self
            }
        }
        if replace_nans_by is not None:
            replace_nans_in_dict(dct, replace_by=replace_nans_by)
        return dct

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)



import pandas
from ..tools import infer_plate_size_from_wellnames
from ..containers import get_plate_class
from ..PickList import PickList, Transfer

def picklist_from_labcyte_echo_picklist_file(filename=None, dataframe=None,
                                             source_plate="auto",
                                             dest_plate="auto"):
    """Return a Picklist from an ECHO picklist file.

    Parameters
    ----------

    filename
      Path to a .csv or .xls(x) file containing a valid ECHO cherrypick list.

    dataframe
      Dataframe obtained by reading a valid ECHO picklist, which can be
      provided instead of ``filename``.

    source_plate, dest_plate
      Plate objects representing the Source and Destination of the picklist.
      If none is provided, new plates will be created, with sizes inferred
      from the content of the picklist.
    """
    if dataframe is None:
        if filename.lower().endswith(".csv"):
            dataframe = pandas.read_csv(filename)
        else:
            dataframe = pandas.read_excel(filename)
    if source_plate == "auto":
        nwells = infer_plate_size_from_wellnames(dataframe["Source Well"])
        source_plate = get_plate_class(nwells)()
        source_plate.name = "Source"
    if dest_plate == "auto":
        nwells = infer_plate_size_from_wellnames(dataframe["Destination Well"])
        dest_plate = get_plate_class(nwells)()
        dest_plate.name = "Destination"
    return PickList([
        Transfer(source_plate.wells[row["Source Well"]],
                 dest_plate.wells[row["Destination Well"]],
                 1e-9 * row["Volume"])
        for i, row in dataframe.iterrows()
    ])



from plateo.exporters import (plate_to_platemap_spreadsheet,
                              plate_to_genesift_sequencing_order_spreadsheet,
                              plate_to_pandas_dataframe,
                              plate_to_bokeh_plot,
                              PlateTextPlotter,
                              PlateGraphsPlotter,
                              PlateColorsPlotter)
from plateo.containers import Plate96

def test_plate_to_platemap_spreadsheet():
    plate = Plate96("TestPlate")


def test_plate_to_genesift_sequencing_order_spreadsheet():
    plate_to_genesift_sequencing_order_spreadsheet


def test_plate_to_pandas_dataframe():
    plate_to_pandas_dataframe


def test_plate_to_bokeh_plot():
    plate_to_bokeh_plot


def test_PlateTextPlotter():
    PlateTextPlotter


def test_PlateGraphsPlotter():
    PlateGraphsPlotter


def test_PlateColorsPlotter():
    PlateColorsPlotter



"""Miscellaneous useful functions.

In particular, methods for converting to and from plate coordinates.
"""

import numpy as np
from collections import OrderedDict
from fuzzywuzzy import process
import re

def compute_rows_columns(num_wells):
    """Convert 96->(8,12), 384->(16,24), etc."""
    a = np.sqrt(num_wells / 6)
    n_rows = int(np.round(2 * a))
    n_columns = int(np.round(3 * a))
    return n_rows, n_columns

def rowname_to_number(name):
    "Convert A->1 Z->26 AA->27 etc."
    if len(name) == 2:
        return 26 * rowname_to_number(name[0]) + rowname_to_number(name[1])
    try:
        return 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.index(name) + 1
    except:
        raise ValueError(name + " is not a valid row name.")


def number_to_rowname(number):
    "Convert 1->A 26->Z 27->AA etc."
    if number > 26:
        return number_to_rowname(int(number / 26)) +\
               number_to_rowname(number % 26)
    return 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'[number - 1]


def wellname_to_coordinates(wellname):
    """Convert A1->(1,1), H11->(8, 11), etc."""
    rowname, colname = re.match("([a-zA-Z]+)([0-9]+)", wellname).groups()
    return rowname_to_number(rowname), int(colname)

def coordinates_to_wellname(coords):
    """Convert (1,1)->A1, (4,3)->D3, (12, 12)->H12, etc."""
    row, column = coords
    return number_to_rowname(row)+str(column)


def wellname_to_index(wellname, num_wells, direction="row"):
    """ Convert e.g. A1..H12 into 1..96
    direction is either row for A1 A2 A3... or column for A1 B1 C1 D1 etc.
    """
    n_rows, n_columns = compute_rows_columns(num_wells)
    row, column = wellname_to_coordinates(wellname)
    if direction == "row":
        return column + n_columns*(row-1)
    elif direction == "column":
        return row + n_rows*(column-1)
    else:
        raise ValueError("`direction` must be in (row, column)")

def index_to_row_column(index, num_wells, direction="row"):
    n_rows, n_columns = compute_rows_columns(num_wells)
    if direction == "row":
        row = 1 + int((index-1) / n_columns)
        column = 1 + ((index-1) % n_columns)
    elif direction == "column":
        row, column = 1 + ((index-1) % n_rows), 1 + int((index-1) / n_rows)
    else:
        raise ValueError("`direction` must be in (row, column)")
    return row, column


def index_to_wellname(index, num_wells, direction="row"):
    """ Convert e.g. 1..96 into A1..H12"""
    row, column = index_to_row_column(index, num_wells, direction)
    return coordinates_to_wellname((row, column))


def shift_wellname(wellname, row_shift=0, column_shift=0):
    letter, number = wellname[0], wellname[1:]
    letter_rownum = rowname_to_number(letter)
    new_letter_rownum = letter_rownum + row_shift
    new_letter = number_to_rowname(new_letter_rownum)
    new_number = str(int(number) + column_shift)
    return new_letter + new_number


def infer_plate_size_from_wellnames(wellnames):
    """Return the first of 96, 384, or 1536, to contain all wellnames."""
    coordinates = [wellname_to_coordinates(name) for name in wellnames]
    all_rows, all_columns = zip(*coordinates)
    max_rows, max_columns = max(all_rows), max(all_columns)
    if (max_rows > 16) or (max_columns > 24):
        return 1536
    elif (max_rows > 8) or (max_columns > 12):
        return 384
    else:
        return 96


def round_at(value, rounding):
    """Round value at the nearest rounding"""
    if rounding is None:
        return value
    else:
        return np.round(value / rounding) * rounding


def dicts_to_columns(dicts):
    return {
        key: [d[key] for d in dicts]
        for key in dicts[0]
    }


def replace_nans_in_dict(dictionnary, replace_by='null'):
    for key, value in dictionnary.items():
        if isinstance(value, dict):
            replace_nans_in_dict(value, replace_by=replace_by)
        elif value == np.nan:
            dictionnary[key] = replace_by

def human_seq_size(n):
    'Return the given sequence as a human friendly 35b, 1.4k, 15k, etc.'
    if n < 1000:
        return '%db' % n
    elif n < 10000:
        return '%.1fk' % (n / 1000)
    else:
        return '%dk' % np.round(n / 1000)

unit_factors = {
    prefix + unit: factor
    for unit in 'glL'
    for prefix, factor in [('', 1), ('m', 1e-3), ('u', 1e-6), ('n', 1e-9)]
}

volume_values_and_units = sorted([
    (value, unit)
    for (unit, value) in unit_factors.items()
    if unit.endswith('L')
])

def find_best_volume_unit(vols):
    med = np.median(vols)
    for value, unit in volume_values_and_units:
        if (not unit.endswith('g')) and (med <= 999 * value):
            return unit
    return unit


def human_volume(vol, unit='auto'):
    if unit == 'auto':
        unit = find_best_volume_unit([vol])
    vol = np.round(vol / unit_factors[unit], 2)
    if int(vol) == vol:
        return "%d %s" % (vol, unit)
    else:
        return "%s %s" % (('%.02f' % vol).rstrip('0'), unit)

def did_you_mean(name, other_names, limit=5, min_score=50):
    if isinstance(name, (list, tuple)):
        return {
            n: did_you_mean(n, other_names, limit=limit, min_score=min_score)
            for n in name
        }
    results = process.extract(name, list(other_names), limit=limit)
    return [e for (e, score) in results if score >= min_score]



from .picklist_from_labcyte_echo_logfile import \
    picklist_from_labcyte_echo_logfile
from .plate_from_tables import plate_from_platemap_spreadsheet

def plate_volumes_from_labcyte_echo_logfile(logfile=None, logcontent=None,
                                            plates_dict=None,
                                            data_field="volume_left"):
    """Return a plate with the volume left in each well after dispenses.

    Parameters
    ----------

    Parameters
    ----------

    logfile
      The path to the echo logfile.

    logcontent
      Echo logfile content, can be provided instead of filename

    plates_dict
      A dictionary of the form {'Plate name': Plate()} linking the plate names
      found in the Echo logs to Plateo Plate objects. If None is provided,
      plates are infered from the Echo logs (a bit experimental).

    data_field
      The well data field that will contain the "volume left" at the end.

    """
    picklist = picklist_from_labcyte_echo_logfile(logfile=logfile,
                                                  logcontent=logcontent)
    source_plates = set([t.source_plate for t in picklist.transfers_list])
    source_plate = list(source_plates)[0]
    for transfer in picklist.transfers_list:
        volume = transfer.data['Current Fluid Volume']*1e-6
        source_plate[transfer.source_well.name].data[data_field] = volume
    return source_plate


def plate_volumes_from_labcyte_echo_survey(filepath=None):
    plate = plate_from_platemap_spreadsheet(filepath, data_field='volume_muL',
                                            skiprows=3)
    def volume(well):
        if 'volume_muL' in well.data:
            return 1e-6 * well.data["volume_muL"]
        else:
            return None

    plate.compute_data_field("volume", volume)
    return plate



from ..PickList import PickList, Transfer
from ..Plate import Plate
from ..containers import Plate96, Plate384, Plate1536
import pandas as pd
import sys
PYTHON3 = (sys.version_info[0] == 3)

if PYTHON3:
    from io import StringIO
else:
    from StringIO import StringIO

def picklist_from_labcyte_echo_logfile(logfile=None, logcontent=None,
                                       plates_dict=None):
    """Return a picklist of what was actually dispensed in the ECHO, based
    on the log file.

    Picklist.metadata["exceptions"] is a picklist of all transfers that went
    wrong.

    Parameters
    ----------

    logfile
      The path to the echo logfile.

    logcontent
      Echo logfile content, can be provided instead of filename

    plates_dict
      A dictionary of the form {'Plate name': Plate()} linking the plate names
      found in the Echo logs to Plateo Plate objects. If None is provided,
      plates are infered from the Echo logs (a bit experimental).

    """

    if plates_dict is None:
        plates_dict = {}

    if logfile is not None:
        with open(logfile) as f:
            logcontent = f.read()

    def read_key_value_block(block, target_dict):
        """Read a block of line of the form 'key, value\nkey, value\n...'"""
        for line in block.splitlines():
            key, value = line.split(",")
            target_dict[key] = value

    def block_to_transfers_list(block):
        """Return a list of transfers from the block of lines."""
        block = "\n".join(block.splitlines()[1:])
        dataframe = pd.read_csv(StringIO(block))
        transfers = []

        for i, row in dataframe.iterrows():
            for role in ("Source", "Destination"):
                plate = row["%s Plate Name" % role]
                if plate not in plates_dict:
                    plate_type = row["%s Plate Type" % role]
                    if "1536" in plate_type:
                        plate_class = Plate1536
                    elif "384" in plate_type:
                        plate_class = Plate384
                    else:
                        plate_class = Plate96
                    plates_dict[plate] = plate_class(
                        name=plate,
                        data={
                            "plate_barcode": row["%s Plate Barcode" % role],
                            "plate_type": plate_type,
                        }
                    )

        for row in dataframe.to_dict(orient="records"):
            source_plate = plates_dict[row["Source Plate Name"]]
            source_well = source_plate.wells[row.pop("Source Well")]
            dest_plate = plates_dict[row["Destination Plate Name"]]
            dest_well = dest_plate.wells[row.pop("Destination Well")]
            volume = float(row.pop("Actual Volume"))
            transfers.append(Transfer(
                volume=(1e-9)*volume,
                source_well=source_well,
                destination_well=dest_well,
                data=row
            ))

        return transfers

    def add_plates_from_metadata(metadata, plates_dict):
        """Auto-find plates in the log file"""
        for role in ("Source", "Destination"):
            plate = metadata["%s Plate Name" % role]
            if plate not in plates_dict:
                plate_type = metadata["%s Plate Type" % role]
                if "1536" in plate_type:
                    num_wells = 1536
                elif "384" in plate_type:
                    num_wells = 384
                else:
                    num_wells = 96
                plates_dict[plate] = Plate(
                    num_wells=num_wells,
                    name=plate,
                    data={
                        "plate_barcode": metadata["%s Plate Barcode" % role],
                        "plate_type": plate_type,
                    }
                )

    blocks = [
        block
        for block in logcontent.split("\r\n\r\n")
        if block != ""
    ]

    picklist_metadata = {"logfile": logfile}
    for block in blocks:

        if block.startswith("Run ID"):
            read_key_value_block(block, target_dict=picklist_metadata)

        if block.startswith("[EXCEPTIONS]"):
            transfers_exceptions_list = block_to_transfers_list(block)
        else:
            transfers_exceptions_list = []

        if block.startswith("[DETAILS]"):
            transfers_list = block_to_transfers_list(block)

        if block.startswith("Instrument Name"):
            read_key_value_block(block, target_dict=picklist_metadata)

    all_transfers = transfers_list + transfers_exceptions_list

    for transfer in all_transfers:
        add_plates_from_metadata(transfer.data, plates_dict)
    for transfer in all_transfers:
        meta = transfer.data
        transfer.source_plate = plates_dict[meta["Source Plate Name"]]
        transfer.destination_plate = plates_dict[
            meta["Destination Plate Name"]]
        obsolete_fields = [
            field for field in meta
            if any(e in field for e in ("Plate Name", "Plate Type",
                                        "Plate Barcode"))
        ]
        for field in obsolete_fields:
            meta.pop(field)

    picklist_metadata["exceptions"] = PickList(transfers_exceptions_list)
    picklist_metadata["plates_dict"] = plates_dict

    return PickList(
        transfers_list=transfers_list,
        data=picklist_metadata
    )



"""Parser for qPCR files.

"""

import pandas as pd
from ..tools import index_to_wellname

def plate_from_roche_lightcycler_qPCR(filename):
    """NOR IMPLEMENTED. TODO: rewrite completely.
    """
    dataframe = pd.read_csv(filename, sep="\t", header=3, index_col=False)
    num_wells = dataframe.No.max()
    return {
        index_to_wellname(row.No, num_wells, direction="row"): {
            column: row[column]
            for column in dataframe.columns
        }
        for i, row in dataframe.iterrows()
    }



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open('plateo/version.py').read()) # loads __version__

setup(name='plateo',
      version=__version__,
      author='Zulko',
    description='Read/write microplate and picklist data for lab automation',
    long_description=open('README.rst').read(),
    license='MIT',
    keywords="microplate biology parser converter report",
    packages= find_packages(exclude='docs'),
    include_package_data=True,
    install_requires= ['pandas>=0.22', 'xlwt', 'xlrd', 'python-box', 'numpy',
                       'matplotlib', 'tqdm', 'pdf_reports', 'flametree',
                       'fuzzywuzzy', 'sequenticon'])



from ..tools import wellname_to_index
import pandas as pd

def optimize_picklist_for_tecan_evo_dispensing(picklist):
    """Return a EVO-optimized version of the picklist.

    In that new version of the Picklist the liquid transfers are sorted, first
    by source plate, then by column of the source well, then by column of the
    destination well.
    """
    return picklist.sorted_by(
        lambda transfer: (transfer.source_well.plate.name,
                          transfer.source_well.column,
                          transfer.destination_well.column)
    )

def picklist_to_tecan_evo_picklist_file(picklist, filename,
                                        change_tips_between_dispenses=True,
                                        optimize_picklist_order=False,
                                        tecan_plate_names=None):
    """

    Parameters
    ----------

    picklist
      The Picklist object to convert.

    filename
      The name in which to write the file. Logically it should have ``.gwl``
      extension (for Gemini Picklist) even though internally it is a ';'-csv

    change_tips_between_dispenses
      If true, a "Wash" step (code W) is added between each transfer.

    optimize_picklist_order
      If true, the picklist output will be optimized for the EVO. This means
      that the liquid transfers are sorted, first by source plate, then by
      column of the source well, then by column of the destination well.

    tecan_plate_names
      A dictionnary ``{plate_object: str_name}`` associating to each plate
      the name it should have in the tecan file.
    """

    if optimize_picklist_order:
        picklist = optimize_picklist_for_tecan_evo_dispensing(picklist)

    tecan_plate_names = {} if tecan_plate_names is None else tecan_plate_names
    def plate_to_tecan_name(plate):
        return tecan_plate_names.get(plate, plate.name)

    columns = [
        "Action", "RackLabel", "RackID", "RackType",
        "Position", "TubeID", "Volume", "LiquidClass",
        "TipType", "TipMask"
    ]

    rows = []
    for transfer in picklist.transfers_list:
        volume = "%.01f" % (transfer.volume / 1e-6)
        absorb = {
            "Action": "A",
            "RackLabel": plate_to_tecan_name(transfer.source_well.plate),
            "Position": wellname_to_index(
                transfer.source_well.name,
                transfer.source_well.plate.num_wells,
                direction="column"
            ),
            "Volume": volume
        }
        dispense = {
            "Action": "D",
            "RackLabel": plate_to_tecan_name(transfer.destination_well.plate),
            "Position": wellname_to_index(
                transfer.destination_well.name,
                transfer.destination_well.plate.num_wells,
                direction="column"
            ),
            "Volume":  volume
        }
        row = [absorb, dispense]
        if change_tips_between_dispenses:
            row += [{"Action": "W"}]
        rows += row

    df = pd.DataFrame.from_records(rows, columns=columns)
    with open(filename, "w+") as f:
        f.write(df.to_csv(sep=";", header=False, index=False).strip("\n"))



from plateo.parsers import (picklist_from_labcyte_echo_logfile,
                            picklist_from_tecan_evo_picklist_file)

def test_picklist_from_labcyte_echo_logfile():
    picklist_from_labcyte_echo_logfile
    pass

def test_picklist_from_tecan_evo_picklist_file():
    picklist_from_tecan_evo_picklist_file
    pass



from plateo.exporters import (
    picklist_to_tecan_evo_picklist_file,
    picklist_to_labcyte_echo_picklist_file,
)


def test_picklist_to_tecan_evo_picklist_file():
    picklist_to_tecan_evo_picklist_file
    pass


def test_picklist_to_labcyte_echo_picklist_file():
    picklist_to_labcyte_echo_picklist_file
